//
//  AmityManager.swift
//  SampleApp
//
//  Created by Michael Abadi on 13/07/20.
//  Copyright © 2020 David Zhang. All rights reserved.
//

import Foundation
import AmitySDK
import Combine

/*
 Shared instance used for convenience in this Sample App.
 
 We recommend having a single instance of AmityClient for your whole app. You can share it using
 singleton or use dependency injection as per your usecase
 */
final class AmityManager: ObservableObject {
    
    static let shared: AmityManager = AmityManager()
    
    private(set) var client: AmityClient?
    
    /// Temporaly saved httpUrl after sdk is initialized
    private var httpUrl: String?
    
    /// An array storing json stirng of push payload, sorted in ascending order.
    @Published private(set) var pushPayloads: [String] = []
    
    private(set) var postRepository: AmityPostRepository?
    
    /// This flag is set at the first page (API Key View Page).
    /// - true: The sample app will login after sdk is initialized.
    /// - false: The sample app will not login after sdk is initialized.
    var needsLogin = true
    
    /// The record of session state history
    @Published private(set) var sessionStateLog: [SessionStateRecord] = []
    
    /// session state observation token.
    private var cancellable: AnyCancellable?
    
    // Production Environment
    // Note:
    // If you want to use default region, you can also use `AmityClient(apiKey: _)` method.
    func setup(apiKey: String, region: AmityRegion) {
        guard !apiKey.isEmpty else {
            assertionFailure("Api key is required!")
            return
        }
        
        Log.add(info: "AmityClient setup with production environment")
        client = try? AmityClient(apiKey: apiKey, region: region)
        httpUrl = region.httpUrl
        setupSessionStateLogger()
        setupRepositories()
    }
    
    // Custom environment
    func setup(environment: ApiEnvironment) {
        let endpoint = AmityEndpoint(httpUrl: environment.httpUrl, mqttHost: environment.mqttHost, uploadUrl: environment.uploadUrl)
        
        Log.add(info: "AmityClient setup with custom environment \(environment.description)")
        client = try? AmityClient(apiKey: environment.apiKey, endpoint: endpoint)
        httpUrl = endpoint.httpUrl
        setupSessionStateLogger()
        setupRepositories()
    }
    
    func setupSessionStateLogger() {
        cancellable = client!.$sessionState.sink { [weak self] sessionState in
            self?.sessionStateLog.append(
                SessionStateRecord(value: sessionState, timestamp: Date())
            )
        }
        
        // To initialize
        NetworkActivityObserver.shared.activities = []
    }
    
    func setupRepositories() {
        guard let client = client else {
            assertionFailure("Client must not be nil at this point.")
            return
        }
        postRepository = AmityPostRepository(client: client)
    }
    
    func appendNotificationPayload(_ userInfo: [AnyHashable: Any]) {
        do {
            let data = try JSONSerialization.data(withJSONObject: userInfo, options: [.prettyPrinted])
            let json = String(data: data, encoding: .utf8)
            pushPayloads.append(json ?? "Unable to convert to json string.")
        } catch {
            Log.add(info: "Unable to convert notification payload to json string.")
        }
    }
    
    func clearNotificationPayloads() {
        pushPayloads = []
    }
    
    var needsAuthToken: Bool {
        if let serverKey = UserDefaults.standard.currentEnvironment?.serverKey {
            return !serverKey.isEmpty
        } else {
            return false
        }
    }
    
    func getAuthToken(userId: String) async throws -> String {
        // prepare parameters
        guard let serverKey = UserDefaults.standard.currentEnvironment?.serverKey else {
            throw GeneralError(message: "unable to find serverKey")
        }
        guard let httpUrl else {
            throw GeneralError(message: "unable to find httpUrl")
        }
        guard var urlComponents = URLComponents(string: httpUrl) else {
            throw GeneralError(message: "unable to find urlComponents")
        }
        // https://_____/api/v3/authentication/token?userId=<user-id>
        urlComponents.path = "/api/v3/authentication/token"
        urlComponents.queryItems = [
            URLQueryItem(name: "userId", value: userId)
        ]
        guard let url = urlComponents.url else {
            throw GeneralError(message: "unable to generate url for getAuthToken")
        }
        // headers = ["x-server-key": "<server-key>"]
        var request = URLRequest(url: url)
        request.addValue(serverKey, forHTTPHeaderField: "x-server-key")
        // send the request
        let (data, response) = try await URLSession.shared.data(for: request)
        // If not 200, throw error.
        let httpResponse = response as! HTTPURLResponse
        guard httpResponse.statusCode == 200 else {
            throw GeneralError(message: "unable to get auth token (\(httpResponse.statusCode)) status code")
        }
        // Can't convert autoken, throw an error.
        guard let authToken = String(data: data, encoding: .utf8) else {
            throw GeneralError(message: "unable to parse auth token")
        }
        // The server return the value as "\"416e9b1ed8dbb3b27178b0d4d147cd979fa88092\""
        // We need to manually remove double-quotes (") at the beginning, and the end of value.
        let output = String(authToken.dropFirst().dropLast())
        // Finally get an auth token
        return output
    }
    
}

struct SessionStateRecord: Identifiable {
    let id = UUID()
    let value: SessionState
    let timestamp: Date
}
