//
//  ReactionTestView.swift
//  SampleApp
//
//  Created by Nishan on 30/8/2566 BE.
//  Copyright © 2566 BE David Zhang. All rights reserved.
//

import SwiftUI
import AmitySDK

struct ReactionTestView: View {
    
    @StateObject var viewModel: ReactionTestViewModel
    
    var body: some View {
        VStack {
            
            HStack {
                Text(viewModel.messageText)
                    .padding()
                    .background(Color(uiColor: .secondarySystemBackground))
                    .cornerRadius(12)
                
                ForEach(viewModel.reactions, id: \.self) { reaction in
                    Image(systemName: reaction == "Like" ? "hand.thumbsup.fill" : "heart.fill")
                        .padding(8)
                        .background(Color(uiColor: .secondarySystemBackground))
                        .clipShape(Circle())
                        .foregroundColor(.green)
                        .accessibilityIdentifier("current-reaction-\(reaction)")
                }
            }
            
            HStack {
                Button {
                    viewModel.replaceReaction(with: "Like")
                } label: {
                    
                    Label("Like", systemImage: "hand.thumbsup.fill")
                        .padding(8)
                        .background(Color(uiColor: .secondarySystemBackground))
                        .cornerRadius(8)
                        .font(.subheadline)
                        .foregroundColor(.black)

                }

                Button {
                    viewModel.replaceReaction(with: "love")
                } label: {
                    
                    Label("Love", systemImage: "heart.fill")
                        .padding(8)
                        .background(Color(uiColor: .secondarySystemBackground))
                        .cornerRadius(8)
                        .font(.subheadline)
                        .foregroundColor(.black)

                }
            }
            .padding(.top)
            
            Text("Tap on like or love to add reaction to this message. It will remove previous reaction and add new reaction.")
                .font(.caption)
        }
    }
}

struct ReactionTestView_Previews: PreviewProvider {
    static var previews: some View {
        ReactionTestView(viewModel: ReactionTestViewModel(messageId: ""))
    }
}

class ReactionTestViewModel: ObservableObject {
    
    var token: AmityNotificationToken?
    let reactionRepo = AmityReactionRepository(client: AmityManager.shared.client!)
    let messageRepo = AmityMessageRepository(client: AmityManager.shared.client!)
    let storyRepo = AmityStoryRepository(client: AmityManager.shared.client!)
    
    enum ReactionType {
        case message
        case story
    }
    
    @Published var messageText: String = ""
    @Published var reactions: [String] = []
    
    let messageId: String
    let type: AmityReactionReferenceType
    
    init(messageId: String, type: AmityReactionReferenceType = .message) {
        self.messageId = messageId
        self.type = type
        switch type {
        case .message:
            self.token = messageRepo.getMessage(messageId).observe({ [weak self] liveObject, error in
                guard let message = liveObject.snapshot else { return }
                
                self?.messageText = message.data?["text"] as? String ?? "-"
                self?.reactions = message.myReactions
            })
        case .story:
                
            self.token = storyRepo.getStory(storyId: messageId).observe({ [weak self] liveObject, error in
                guard let story = liveObject.snapshot else { return }
                
                self?.messageText = story.storyId
                self?.reactions = story.myReactions
            })
        default:
            break
        }
        
    }
    
    func replaceReaction(with reactionName: String) {
        guard !reactions.contains(reactionName) else { return } // If its already there, do nothing
        let shouldReplace = !reactions.isEmpty
        let reactionToRemove = reactionName == "Like" ? "love" : "Like"
        
        if shouldReplace {
            removeReaction(reaction: reactionToRemove, messageId: messageId) { [weak self] success in
                guard let self else { return }
                
                self.addReaction(reaction: reactionName, messageId: self.messageId) { success in
                    // Added...
                }
            }
        } else {
            addReaction(reaction: reactionName, messageId: messageId) { success in
                
            }
        }
    }
    
    private func addReaction(reaction: String, messageId: String, completion: @escaping (Bool) -> Void) {
        Task {
            do {
                let status = try await reactionRepo.addReaction(reaction, referenceId: messageId, referenceType: type)
                completion(status)
            } catch let error {
                completion(false)
            }
        }
    }
    
    private func removeReaction(reaction: String, messageId: String, completion: @escaping (Bool) -> Void) {
        Task {
            do {
                let status = try await reactionRepo.removeReaction(reaction, referenceId: messageId, referenceType: type)
                completion(status)
            } catch let error {
                completion(false)
            }
        }
    }
}
