//
//  ApiKeyView.swift
//  SampleApp
//
//  Created by Nishan Niraula on 10/8/20.
//  Copyright © 2020 David Zhang. All rights reserved.
//

import SwiftUI
import AmitySDK
import Combine

/*
 This class is for sample app purpose only. This is not a production ready code.
 Please don't use it in production environment.
 */

enum EnvironmentType: String, CaseIterable, Codable {
    case production
    case staging
    case local
    case custom
    
    var title: String {
        switch self {
        case .production:
            return "Production"
        case .staging:
            return "Staging"
        case .local:
            return "Local"
        case .custom:
            return "Custom"
        }
    }
}

struct ApiKeyView: View {

    @State private var environments: [ApiEnvironment]
    @State private var selectedEnvironment: ApiEnvironment
    @State private var selectedEnvironmentId: String = ""

    @State private var selectedRegion = AmityRegion.SG
    
    @State private var enteredApiKey: String
    @State private var enteredUserId: String
    @State private var enteredHttpUrl: String
    @State private var enteredMqttHost: String
    @State private var enteredUploadUrl: String
    
    @State private var enabledAuthKey: Bool
    @State private var enteredServerKey: String
    
    @Binding var isSetupComplete: Bool
    @State private var showingAlert = false
    
    init(isSetupComplete: Binding<Bool>) {
        _isSetupComplete = isSetupComplete
        
        // Environments Setup
        var savedEnvironments = UserDefaults.standard.getSavedEnvironments()
        if savedEnvironments.isEmpty {
            // If there is no data, save default environments.
            savedEnvironments = ApiEnvironment.defaultEnvironments
            UserDefaults.standard.saveEnvironments(savedEnvironments)
        }
        
        // Retrieve selected env
        var currentEnvironment: ApiEnvironment
        if let currentEnvironmentId = UserDefaults.standard.currentEnvironmentId,
           let index = savedEnvironments.firstIndex(where: { $0.id == currentEnvironmentId }) {
            currentEnvironment = savedEnvironments[index]
        } else {
            // current env doesn't set yet. pick the first value then save to UserDefaults.
            currentEnvironment = savedEnvironments.first!
            UserDefaults.standard.currentEnvironmentId = currentEnvironment.id
        }
        
        _environments = State(initialValue: savedEnvironments)
        _selectedEnvironment = State(initialValue: currentEnvironment)
        _selectedEnvironmentId = State(initialValue: currentEnvironment.id)
        _enteredHttpUrl = State(initialValue: currentEnvironment.httpUrl)
        _enteredMqttHost = State(initialValue: currentEnvironment.mqttHost)
        _enteredApiKey = State(initialValue: currentEnvironment.apiKey)
        _enteredUserId = State(initialValue: UserDefaults.standard.userId ?? "victimIOS")
        
        _enabledAuthKey = State(initialValue: !currentEnvironment.serverKey.isEmpty)
        _enteredServerKey = State(initialValue: currentEnvironment.serverKey)
        _enteredUploadUrl = State(initialValue: currentEnvironment.uploadUrl)
    }
    
    var body: some View {
        NavigationView {
            Form {
                
                // Environment Setup & Selection
                Section(header: Text("Setup Environment")) {
                    NavigationLink(destination: ApiEnvironmentView(onCreate: { environment in
                        environments.append(environment)
                    })) {
                        Text("Add New Environment")
                            .font(.subheadline)
                    }
                    
                    Picker(selection: $selectedEnvironmentId, label: Text("Selected Environment")) {
                        ForEach(environments, id: \.id) { env in
                            Text(env.name)
                                .lineLimit(2)
                                .font(.caption)
                        }
                    }.onChange(of: selectedEnvironmentId) { environmentId in
                        guard !selectedEnvironmentId.isEmpty,
                              let index = environments.firstIndex(where: { $0.id == environmentId }) else { return }
                        selectedEnvironment = environments[index]
                        enteredApiKey = selectedEnvironment.apiKey
                        enteredHttpUrl = selectedEnvironment.httpUrl
                        enteredMqttHost = selectedEnvironment.mqttHost
                        enteredUploadUrl = selectedEnvironment.uploadUrl
                    }
                }
                
                // Region setup for production environment
                Section(header: Text("Api Key & Region")) {
                    TextField("Api Key", text: $enteredApiKey)
                        .font(.footnote)
                    
                    if selectedEnvironment.envType == .production {
                        Picker("Region:", selection: $selectedRegion) {
                            let regions: [AmityRegion] = [.EU, .SG, .US]
                            ForEach(regions, id: \.self) { region in
                                Text(region.regionIdentifier).tag(region.rawValue)
                            }
                        }
                    }
                }
                
                if selectedEnvironment.envType != .production {
                    Section(header: Text("Current Selection:")) {
                        Group {
                            TextField("Http URL", text: $enteredHttpUrl)
                            TextField("Mqtt Host", text: $enteredMqttHost)
                            TextField("Upload URL", text: $enteredUploadUrl)
                        }
                        .font(.subheadline)
                    }
                }
                Section {
                    Toggle(isOn: $enabledAuthKey) {
                        Text("Use auth token when login")
                    }
                    if enabledAuthKey {
                        TextField("<server-key>", text: $enteredServerKey)
                            .font(.subheadline)
                            .autocapitalization(.none)
                    }
                } header: {
                    Text("Secure Mode")
                } footer: {
                    Text("When enabled the sample app will automatically get an auth token for login / renew.")
                }
                
                Section(header: Text("User Id")) {
                    TextField("User Id:", text: $enteredUserId)
                        .font(.subheadline)
                        .autocapitalization(.none)
                }
                
                Section(header: Text("Establish new login session")) {
                    FormButton(title: "Initialize SDK and login") {
                        initializeSDK(needsLogin: true)
                    }
                }

                Section(header: Text("Resume previously logged in session")) {
                    FormButton(title: "Initialize SDK only") {
                        initializeSDK(needsLogin: false)
                    }
                }
                
            }
            .navigationBarTitle("SDK Setup", displayMode: .inline)
            .toolbar {
                ToolbarItem(placement: .navigationBarTrailing) {
                    Button("Reset ENV") {
                        environments = ApiEnvironment.defaultEnvironments
                        selectedEnvironment = environments.first!
                        selectedEnvironmentId = selectedEnvironment.id
                        UserDefaults.standard.saveEnvironments(environments)
                        UserDefaults.standard.currentEnvironmentId = selectedEnvironmentId
                        UserDefaults.standard.synchronize()
                    }
                }
            }
        }
    }
    
    private func initializeSDK(needsLogin: Bool) {
        guard !enteredUserId.isEmpty else { return }
        
        // Amity Client Setup
        // !!! IMPORTANT !!!
        //
        // SDK needs to be setup before any method can be used. This is a two step process.
        //
        // Step 1: Create an instance of AmityClient with provided apikey.
        // This step is handled in AmityManager class.
        
        // Step 2: Register userId using `login(_:)` method.
        // This step is handled in HomeViewController class.
        
        // Save data changes to UserDefaults.
        var updatedEnv = selectedEnvironment
        updatedEnv.httpUrl = enteredHttpUrl
        updatedEnv.mqttHost = enteredMqttHost
        updatedEnv.apiKey = enteredApiKey
        updatedEnv.uploadUrl = enteredUploadUrl
        if !enabledAuthKey {
            updatedEnv.serverKey = ""
        } else {
            updatedEnv.serverKey = enteredServerKey
        }
        
        UserDefaults.standard.updateEnvironment(updatedEnv)
        
        selectedEnvironment = updatedEnv
        
        UserDefaults.standard.currentEnvironmentId = selectedEnvironmentId
        UserDefaults.standard.userId = enteredUserId
        UserDefaults.standard.synchronize()
        
        AmityManager.shared.needsLogin = needsLogin
        
        switch selectedEnvironment.envType {
        case .production:
            AmityManager.shared.setup(apiKey: selectedEnvironment.apiKey, region: selectedRegion)
        case .staging, .local, .custom:
            AmityManager.shared.setup(environment: selectedEnvironment)
        }
        
        isSetupComplete = true
    }
    
}

struct ApiKeyView_Previews: PreviewProvider {
    
    static var previews: some View {
        Text("ApiKey View")
    }
}

struct MainContainerView: View {
    @State var isSetupComplete = false
    
    var body: some View {
        if self.isSetupComplete {
            if #available(iOS 14.0, *) {
                return AnyView(SwiftUIMainTabView().ignoresSafeArea(.keyboard, edges: .bottom))
            } else {
                return AnyView(SwiftUIMainTabView())
            }
        } else {
            return AnyView(ApiKeyView(isSetupComplete: $isSetupComplete))
        }
    }
}

extension AmityRegion: Codable { }

struct ApiEnvironment: Codable {
    
    var id: String = UUID().uuidString
    var name: String = ""
    var apiKey: String = ""
    var region: AmityRegion = .SG
    var httpUrl: String = ""
    var mqttHost: String = ""
    var serverKey: String = ""
    var uploadUrl: String = ""
    
    var envType: EnvironmentType = .custom
    
    func isValid() -> Bool {
        guard !name.isEmpty && !apiKey.isEmpty && !httpUrl.isEmpty else { return false }
        return true
    }
    
    static func encodeData(data: ApiEnvironment) -> Data? {
        let encodedData = try? JSONEncoder().encode(data)
        return encodedData
    }
    
    static func decodeData(data: Data) -> ApiEnvironment? {
        let decodedData = try? JSONDecoder().decode(ApiEnvironment.self, from: data)
        return decodedData
    }
    
    var description: String {
        return "Name: \(name) ApiKey: \(apiKey), Http: \(httpUrl), Mqtt: \(mqttHost), serverKey: \(serverKey)"
    }
    
    static var defaultEnvironments: [ApiEnvironment] {
        
        // prod
        let prodEnv = ApiEnvironment(id: "production-env",
                                     name: EnvironmentType.production.title,
                                     apiKey: "b3babb5e32dbf2314935d81e0301438fd85b84b7b2316a2e",
                                     region: .SG,
                                     envType: .production)
        
        // staging
        let stagingEnv = ApiEnvironment(id: "staging-env",
                                        name: EnvironmentType.staging.title,
                                        apiKey: "b0efe90c3bdda2304d628918520c1688845889e4bc363d2c",
                                        httpUrl: "https://api.staging.amity.co",
                                        mqttHost: "ssq.staging.amity.co",
                                        uploadUrl: "https://upload.staging.amity.co",
                                        envType: .staging)
        
        // local03
        let localEnv = ApiEnvironment(id: "local-env",
                                      name: EnvironmentType.local.title,
                                      apiKey: "b0eeeb093bdff2621d348c495b591eded85f8ce3ee326779",
                                      httpUrl: "https://asc03.eko.ninja:8443",
                                      mqttHost: "ssq.staging.amity.co",
                                      uploadUrl: "https://asc03.eko.ninja:8443",
                                      envType: .local)
        
        return [prodEnv, stagingEnv, localEnv]
    }
    
}

struct ApiEnvironmentView: View {
    
    @State private(set) var environment: ApiEnvironment = ApiEnvironment()
    @State private var buttonText = "Create"
    
    @Environment(\.presentationMode) var presentation
    
    var onCreate: (ApiEnvironment) -> Void
    
    var hint = """
    Format:
    
    Http Url: https://api.amity.co
    Socket Url: https://api.amity.co
    Mqtt Host: ssq.amity.co (without scheme mqtts://)
    
    All fields are mandatory!
    """
    
    var body: some View {
        Form {
            Section(header: Text("Add New Environment"), footer: Text(hint).font(.footnote)) {
                TextField("Environment Name:", text: $environment.name)
                TextField("Api Key:", text: $environment.apiKey)
                    .autocapitalization(.none)
                TextField("Http Url:", text: $environment.httpUrl)
                    .autocapitalization(.none)
                TextField("Mqtt Host", text: $environment.mqttHost)
                    .autocapitalization(.none)
            }
            .font(.subheadline)
            
            Section {
                HStack {
                    Spacer()
                    Button {
                        guard environment.isValid() else { return }
                        
                        UserDefaults.standard.addEnvironment(environment)
                        UserDefaults.standard.synchronize()
                        
                        onCreate(environment)
                        
                        // Pop screen
                        presentation.wrappedValue.dismiss()
                        
                    } label: {
                        Text(buttonText)
                    }
                    Spacer()
                }
            }
        }
        .onAppear {
            if !environment.name.isEmpty {
                buttonText = "Update"
            } else {
                buttonText = "Create"
            }
        }
    }
}

struct FormButton: View {
    
    let title: String
    let action: (() -> Void)?
    
    var body: some View {
        Button {
            action?()
        } label: {
            HStack {
                Spacer()
                Text(title)
                Spacer()
            }
        }
    }
}
