//
//  GetStoryTargetsView.swift
//  SampleApp
//
//  Created by Manuchet Rungraksa on 12/1/2567 BE.
//  Copyright © 2567 BE David Zhang. All rights reserved.
//

import Foundation
import AmitySDK
import SwiftUI

struct GetStoryTargetsView: View {
    
    @ObservedObject var viewModel: StoryTargetListViewModel
    
    var body: some View {
        VStack(alignment: .center) {
            Text("Format is \"targetType/targetId\" per search target and separate by \", \" (comma and space)")
                .font(.caption)
            TextField("Search by targetType/targetId", text: $viewModel.textTarget)
                .textFieldStyle(RoundedBorderTextFieldStyle())
                .font(.callout)
            
            
            Button(action: {
                viewModel.queryStoryTargets()
            }){
                Text("Query story list")
                    .frame(alignment: .center)
            }.buttonStyle(BorderlessButtonStyle())
            
            if let errorString = viewModel.errorString {
                Text(errorString)
            }
            List(viewModel.storyTargets) { target in
                VStack(alignment: .leading) {
                    Text("Target Id: \(target.targetId)").frame(alignment: .leadingFirstTextBaseline)
                    Text("Target type: \(target.targetType)")
                    Text("Has Unseen: \(target.hasUnseen)" as String)
                }
            }
        }
    }
}

class StoryTargetListViewModel: ObservableObject {
    
    var pageTitle: String
    
    @Published var textTarget: String = ""
    @Published var errorString: String?
        
    private var token: AmityNotificationToken?
    private var storyTargetCollection: AmityCollection<AmityStoryTarget>?
    private var storyRepository = AmityStoryRepository(client: AmityManager.shared.client!)
    private var shouldFetchNextPage = false
            
    @Published var storyTargets: [StoryTargetListModel] = []
    
    init() {
        self.pageTitle = "User list"
        
    }
    
    func fetchNextPage() {
        guard storyTargetCollection?.hasNext ?? false else {
            shouldFetchNextPage = true
            return
        }
        storyTargetCollection?.nextPage()
    }
    
    func queryStoryTargets() {
        var searchTargets = [AmityStoryTargetSearchInfo]()
        let eachTarget = textTarget.components(separatedBy: ", ")
        for target in eachTarget {
            let splitedText = target.components(separatedBy: "/")
            let type = splitedText.first ?? ""
            let id = splitedText.last ?? ""
            
            let searchTarget = AmityStoryTargetSearchInfo(targetType: AmityStoryTargetType(rawValue: type) ?? .community, targetId: id)
            searchTargets.append(searchTarget)
        }
        
        storyTargets.removeAll()
        errorString = nil
        token?.invalidate()
        
        storyTargetCollection = storyRepository.getStoryTargets(targets: searchTargets)
        token = storyTargetCollection?.observe({ [weak self] collection, _, error in
            
            if let error {
                self?.errorString = error.localizedDescription
            }
            
            guard let self = self else { return}
            
            if collection.dataStatus == .fresh, self.shouldFetchNextPage {
                shouldFetchNextPage = false
                self.storyTargetCollection?.nextPage()
            }
            var targets = [StoryTargetListModel]()
            for storyTarget in collection.snapshots {
                let model = StoryTargetListModel(storyTarget: storyTarget)
                targets.append(model)
            }
            self.storyTargets = targets
            
        })
        
    }

    @MainActor
    func queryGlobalFeedStoryTargets(option: AmityGlobalStoryTargetsQueryOption) {
        
        storyTargets.removeAll()
        errorString = nil
        token?.invalidate()
        
        storyTargetCollection = storyRepository.getGlobalStoryTargets(option: option)
        
        token = storyTargetCollection?.observe({ [weak self] collection, _, error in

            if let error {
                self?.errorString = error.localizedDescription
            }
            
            guard let self = self else { return}
            
            if collection.dataStatus == .fresh, self.shouldFetchNextPage {
                shouldFetchNextPage = false
                self.storyTargetCollection?.nextPage()
            }
            
            var targets = [StoryTargetListModel]()
            for storyTarget in collection.snapshots {
                let model = StoryTargetListModel(storyTarget: storyTarget)
                targets.append(model)
            }
            self.storyTargets = targets
            
        })
        
    }
    
}

struct StoryTargetListModel: Identifiable {
    var id: String
    
    let storyTarget: AmityStoryTarget
    
    let targetType: String
    let targetId: String
    let hasUnseen: Bool
    let failedStoriesCount: Int
    let syncingStoriesCount: Int
    let community: AmityCommunity?
    
    init(storyTarget: AmityStoryTarget) {
        self.storyTarget = storyTarget        
        self.targetType = storyTarget.targetType
        self.targetId = storyTarget.targetId
        self.id = storyTarget.targetId
        self.hasUnseen = storyTarget.hasUnseen
        self.failedStoriesCount = storyTarget.failedStoriesCount
        self.syncingStoriesCount = storyTarget.syncingStoriesCount
        self.community = storyTarget.community
    }
}
