//
//  CommunitySemanticSearchView.swift
//  SampleApp
//
//  Created by Nishan Niraula on 27/11/24.
//  Copyright © 2024 David Zhang. All rights reserved.
//

import SwiftUI
import AmitySDK

struct CommunitySemanticSearchView: View {
    
    @State private var categoryIds: String = ""
    @State private var membershipStatus: Int = 0 // 2 means non member
    @State private var tags: String = ""
    @State private var query: String = ""
    @State private var showSearchSheet = false
    
    @StateObject var viewModel: CommunitySemanticSearchViewModel = CommunitySemanticSearchViewModel()
    
    var body: some View {
        VStack {
            
            List {
                ForEach(viewModel.searchedCommunities, id: \.communityId) { community in
                    VStack(alignment: .leading) {
                        Text(community.modelDescription)
                    }
                    .font(.system(size: 15))
                    .foregroundColor(Color(UIColor.label))
                    .onAppear {
                        if let lastItem = viewModel.searchedCommunities.last, lastItem.communityId == community.communityId {
                            
                            viewModel.loadMoreCommunities()
                        }
                    }
                }
            }
            
            AmityDefaultButton(title: "Search Communities") {
                showSearchSheet.toggle()
            }
            .alert("", isPresented: $viewModel.showAlert) {
                Button {
                    viewModel.showAlert = false
                } label: {
                    Text("Dismiss")
                }
            } message: {
                Text(viewModel.alertMessage)
            }
            
        }.sheet(isPresented: $showSearchSheet, content: {
            NavigationView {
                ScrollView {
                    VStack(alignment: .leading, spacing: 12) {
                        
                        Text("Search Query:")
                            .asFormLabel()
                            .padding(.top, 16)
                        
                        TextField("Enter Search Query", text: $query)
                            .modifier(AmityTextFieldStyle())
                        
                        Text("Category Ids (optional):")
                            .asFormLabel()
                        
                        TextField("Ids:", text: $categoryIds)
                            .modifier(AmityTextFieldStyle())
                        
                        Text("Enter multiple category ids by separating with comma. Ex, abc,def,ijk")
                            .asDescription()
                        
                        Text("Tags (optional):")
                            .asFormLabel()
                        
                        TextField("tags:", text: $tags)
                            .modifier(AmityTextFieldStyle())
                        
                        Text("Enter multiple tags by separating with comma. Ex, abc,def,ijk")
                            .asDescription()
                        
                        Text("Membership Status")
                            .asFormLabel()

                        Picker("Target Type", selection: $membershipStatus) {
                            Text("All")
                                .tag(0)
                            Text("Member")
                                .tag(1)
                            Text("Not Member")
                                .tag(2)
                        }
                        .pickerStyle(.segmented)
                                        
                    }
                    .padding(.horizontal)
                }
                .navigationTitle("Search Options")
                .navigationBarTitleDisplayMode(.inline)
                .toolbar(content: {
                    ToolbarItem(placement: .primaryAction) {
                        Button(action: {
                            
                            let catIds = self.categoryIds.components(separatedBy: ",").map { $0.trimmingCharacters(in: .whitespacesAndNewlines) }.filter { !$0.isEmpty }
                            let tags = self.tags.components(separatedBy: ",").map { $0.trimmingCharacters(in: .whitespacesAndNewlines)}.filter { !$0.isEmpty }
                            
                            var membership: AmityCommunityMemberStatusFilter = .all
                            
                            if membershipStatus == 0 {
                                membership = .all
                            } else if membershipStatus == 1 {
                                membership = .member
                            } else if membershipStatus == 2 {
                                membership = .notMember
                            }
                            
                            let searchOptions = AmityCommunitySemanticSearchOptions(query: query, categoryIds: catIds.isEmpty ? nil : catIds, tags: tags.isEmpty ? nil : tags, communityMembershipStatus: membership)
                            viewModel.searchCommunities(options: searchOptions)
                            
                            showSearchSheet = false
                        }, label: {
                            Text("Search")
                        })
                    }
                })
            }
            .modifier(HalfModal())
        })
    }
    
    struct HalfModal: ViewModifier {
        
        func body(content: Content) -> some View {
            if #available(iOS 16.0, *) {
                content
                    .presentationDetents([.fraction(0.68), .large])
            } else {
                content
            }
        }
    }
}

#Preview {
    CommunitySemanticSearchView()
}

class CommunitySemanticSearchViewModel: ObservableObject {
    
    @Published var searchedCommunities: [AmityCommunity] = []
    @Published var showAlert = false
    @Published var alertMessage = ""
    
    let commRepo = AmityCommunityRepository(client: AmityManager.shared.client!)
    var token: AmityNotificationToken?
    var collection: AmityCollection<AmityCommunity>?
    
    func searchCommunities(options: AmityCommunitySemanticSearchOptions) {
        token?.invalidate()
        token = nil
        
        searchedCommunities = []
        
        collection = commRepo.semanticSearchCommunities(options: options)
        token = collection?.observe { liveCollection, _, error in
            
            if let error {
                self.showAlert = true
                self.alertMessage = error.localizedDescription
                self.searchedCommunities = []
                return
            }
            
            self.searchedCommunities = liveCollection.allObjects()
        }
    }
    
    func loadMoreCommunities() {
        if let collection, collection.hasNext {
            collection.nextPage()
        }
    }
}
